import django
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import permission_required
from django.core.exceptions import PermissionDenied
from django.views.generic import DetailView


@staff_member_required
def view_needing_staff(request):
    # If this line is reached, the test as failed, as the
    # view visibility code should not execute views
    raise Exception  # pragma: no cover


@permission_required('auth.add_user')
def view_needing_add_user(request):
    # If this line is reached, the test as failed, as the
    # view visibility code should not execute views
    raise Exception  # pragma: no cover


class ViewNeedingAddUser(DetailView):
    def dispatch(self, request, *args, **kwargs):
        if request.user.has_perm('auth.add_user'):
            return super(ViewNeedingAddUser, self).dispatch(request, *args, **kwargs)
        else:
            raise PermissionDenied

    def get(self, *_args, **_kwargs):
        # If this line is reached, the test as failed, as the
        # view visibility code should not execute views
        raise Exception  # pragma: no cover

if django.VERSION >= (1, 9, 0):
    from django.contrib.auth.mixins import PermissionRequiredMixin

    class MixinViewNeedingAddUser(PermissionRequiredMixin, DetailView):
        permission_required = 'auth.add_user'
        raise_exception = True

        def get(self, *_args, **_kwargs):
            # If this line is reached, the test as failed, as the
            # view visibility code should not execute views
            raise Exception  # pragma: no cover



'''Test URLs'''
from django.conf.urls import patterns

urlpatterns = patterns('',
    (r'^error400/$', 'nuit.handlers.handler400'),
    (r'^error403/$', 'nuit.handlers.handler403'),
    (r'^error404/$', 'nuit.handlers.handler404'),
    (r'^error500/$', 'nuit.handlers.handler500'),
)



from django import forms

class TestForm(forms.Form):
    title = forms.CharField()
    firstname = forms.CharField()
    lastname = forms.CharField()
    email = forms.CharField()
    url = forms.URLField()
    message = forms.CharField()


from __future__ import division
from django import template
from django.contrib.messages import constants
from django.core.urlresolvers import resolve, reverse, NoReverseMatch, Resolver404
from django.template.loader import get_template
from django.template.base import token_kwargs, FilterExpression
from django.template.loader_tags import do_extends, ExtendsNode
from django.template.defaultfilters import slugify
from django.utils.html import format_html
from ast import literal_eval
from six.moves.urllib.parse import urlparse
from six import iteritems as _iteritems
from ..utils import user_can_see_view

# pylint: disable=C0103

register = template.Library()

NoneFilterExpression = FilterExpression("None", None)
FalseFilterExpression = FilterExpression("False", None)
TrueFilterExpression = FilterExpression("True", None)

MESSAGE_LEVELS = {
    constants.ERROR: 'alert',
}

@register.filter
def message_class(msg):
    '''
    Return the foundation alert class for a message level.
    '''
    try:
        return MESSAGE_LEVELS[msg.level]
    except KeyError:
        try:
            return constants.DEFAULT_TAGS[msg.level]
        except KeyError:
            return msg.level

@register.simple_tag
def set_active_menu(active_menu):
    '''
    Inserts a span with a class of .nuit-active-menu that is picked up by Javascript
    to highlight the correct menu item.
    '''
    return format_html("<span class='hide nuit-active-menu'>{}</span>", active_menu)


class ExtendNode(ExtendsNode):
    '''
    Template node that extends another template with additional variables.
    '''

    def __init__(self, node, kwargs):
        super(ExtendNode, self).__init__(node.nodelist, node.parent_name, node.template_dirs)
        self.kwargs = dict(("nuit_%s" % key, value) for key, value in _iteritems(kwargs))

    def __repr__(self):
        return '<ExtendNode: extends %s with args: %r>' % (super(ExtendNode, self).__repr__(), self.kwargs)

    def render(self, context):
        kwargs = dict((key, value.resolve(context)) for key, value in _iteritems(self.kwargs) if key not in context)
        context.update(kwargs)
        try:
            return super(ExtendNode, self).render(context)
        finally:
            context.pop()

@register.tag
def extend(parser, token):
    '''
    Extends another template, passing additional variables to the parent.
    '''
    bits = token.split_contents()
    kwargs = token_kwargs(bits[2:], parser)
    token.contents = " ".join(bits[:2])
    return ExtendNode(do_extends(parser, token), kwargs)


class MenuSectionNode(template.Node):
    '''
    Template node that renders a menu section.
    '''
    # pylint: disable=W0622

    def __init__(self, nodelist, title=NoneFilterExpression, is_list=FalseFilterExpression, link_name=NoneFilterExpression, id=NoneFilterExpression, can_collapse=FalseFilterExpression):
        self.nodelist = nodelist
        self.title = title
        self.is_list = is_list
        self.link_name = link_name
        self.link_id = id
        self.can_collapse = can_collapse

    def render(self, context):
        content = self.nodelist.render(context)
        bare_title = self.title.resolve(context)
        title = ''
        if bare_title:
            title = '<h5>%s%s</h5>' % (bare_title, '<i class=\'nuit-collapse fi-play right\'></i>' if self.can_collapse.resolve(context) else '')
        link_name = self.link_name.resolve(context) or bare_title
        resolved_link_id = self.link_id.resolve(context)
        return '''
            <section class='right-menu-reveal' {reveal} data-link='{link_name}' id='{id}'>
                <a class='close-reveal-modal'>&#215;</a>
                <div>
                {title}
                <div class='side-content'>
                {list_begin}
                {content}
                {list_end}
                </div>
                </div>
                <hr />
            </section>
        '''.format(
            title = title,
            content = content,
            link_name = link_name,
            id = resolved_link_id or slugify(link_name),
            list_begin = "<nav><ul class='side-nav'>" if self.is_list.resolve(context) else '',
            list_end = '</ul></nav>' if self.is_list.resolve(context) else '',
            reveal = 'data-reveal' if not self.is_list.resolve(context) else '',
        )



@register.tag
def menu_section(parser, token):
    '''
    Renders the correct HTML for a menu section.
    '''
    args = token.split_contents()
    kwargs = {}
    for kwarg in args[1:]:
        if '=' not in kwarg:
            kwargs['title'] = kwarg
            continue
        before, after = kwarg.split('=')
        kwargs[before] = after
    nodelist = parser.parse(('end_menu_section',))
    parser.delete_first_token()
    kwargs = dict((key, parser.compile_filter(value)) for key, value in _iteritems(kwargs))
    return MenuSectionNode(nodelist, **kwargs)


class AppMenuNode(template.Node):
    '''
    Template node that renders the main application menu for Nuit.
    '''

    def __init__(self, nodelist, title=NoneFilterExpression, can_collapse=FalseFilterExpression):
        self.nodelist = nodelist
        self.title = title
        self.can_collapse = can_collapse

    def render(self, context):
        content = self.nodelist.render(context)
        title = '<h5>%s%s</h5>' % (self.title.resolve(context) if self.title else '', '<i class=\'nuit-collapse fi-play right\'></i>' if self.can_collapse.resolve(context) else '')
        return "<section class='main-nav'>{title}<nav><ul class='side-content side-nav'>{content}</ul></nav><hr /></section>".format(title=title, content=content)

@register.tag
def app_menu(parser, token):
    '''
    Renders the main application menu for Nuit.
    '''
    args = token.split_contents()
    kwargs = {}
    for kwarg in args[1:]:
        if '=' not in kwarg:
            kwargs['title'] = kwarg
            continue
        before, after = kwarg.split('=')
        kwargs[before] = after
    nodelist = parser.parse(('end_app_menu',))
    parser.delete_first_token()
    kwargs = dict((key, parser.compile_filter(value)) for key, value in _iteritems(kwargs))
    return AppMenuNode(nodelist, **kwargs)


@register.simple_tag(takes_context=True)
def menu_item(context, link, name, id=None, current=False, unavailable=False, always_display=False):
    '''
    Renders an HTML anchor element in a list element.
    '''
    # pylint: disable=W0622
    if not id:
        id = slugify(name)
    classes = []
    if current:
        classes.append('current')
    if unavailable:
        classes.append('unavailable')

    try:
        url = reverse(link)
    except NoReverseMatch:
        url = link
        display = True

    parsed_url = urlparse(url)

    if not parsed_url.netloc and not always_display:
        user = context['request'].user if 'request' in context else None
        try:
            view = resolve(parsed_url.path)
        except Resolver404:
            display = False
        else:
            display = user_can_see_view(view, user)
    else:
        display = True

    if always_display or display:
        return format_html("<li class='menu-item menu-{id} {classes}'><a class='menu-item' href='{link}'>{name}</a></li>", name=name, link=url, id=id, classes=' '.join(classes))
    else:
        return ''


@register.inclusion_tag('nuit/includes/_pagination_menu.html', takes_context=True)
def pagination_menu(context, page_obj, show_totals=True):
    '''
    Renders the HTML required for the pagination options.
    '''
    total_pages = page_obj.paginator.page_range
    actual_numbers = [page for page in total_pages if page >= total_pages[-1] - 1 or page <= 2 or (page >= page_obj.number - 2 and page <= page_obj.number + 2)]
    page_list = []
    for i, number in enumerate(actual_numbers[:-1]):
        page_list.append(number)
        if actual_numbers[i + 1] != number + 1:
            page_list.append(None)
    page_list.append(actual_numbers[-1])
    return {
        'context': context,
        'page_obj': page_obj,
        'page_list': page_list,
        'show_totals': show_totals,
    }


def calculate_widths(parts, num=12):
    '''
    Generates a list of length parts containing integers, the sum of which is num.
    '''
    first = num // parts
    last  = num - (parts - 1) * first
    return (parts - 1) * [first] + [last]


class FoundationFormField(object):
    '''
    A form field, containing Foundation-markup specific parameters.
    '''
    # pylint: disable=R0902
    # pylint: disable=C0301

    def __init__(self, field, small=12, medium=None, large=None, prefix=None, prefix_small=3, prefix_medium=None, prefix_large=None, postfix=None, postfix_small=3, postfix_medium=None, postfix_large=None, show_label=True, label=None, switch=False, switch_class=""):
        # pylint: disable=R0913
        self.field = field
        self.small_width = small
        self.medium_width = medium or self.small_width
        self.large_width = large or self.medium_width
        self.prefix = prefix
        self.prefix_small_width = prefix_small
        self.prefix_medium_width = prefix_medium or self.prefix_small_width
        self.prefix_large_width = prefix_large or self.prefix_medium_width
        self.postfix = postfix
        self.postfix_small_width = postfix_small
        self.postfix_medium_width = postfix_medium or self.postfix_small_width
        self.postfix_large_width = postfix_large or self.postfix_medium_width
        self.show_label = show_label
        self.label = label
        self.switch = switch
        self.switch_class = switch_class

        if self.prefix or self.postfix:
            self.fix_field_small_width = 12 - (self.prefix_small_width if self.prefix else 0) - (self.postfix_small_width if self.postfix else 0)
            self.fix_field_medium_width = 12 - (self.prefix_medium_width if self.prefix else 0) - (self.postfix_medium_width if self.postfix else 0)
            self.fix_field_large_width = 12 - (self.prefix_large_width if self.prefix else 0) - (self.postfix_large_width if self.postfix else 0)

def normalise_row(row_data):
    # pylint: disable=R0912
    sizes = {
        'field': ('small', 'medium', 'large',)
    }
    sizes['prefixes'] = tuple('prefix_%s' % s for s in sizes['field'])
    sizes['postfixes'] = tuple('postfix_%s' % s for s in sizes['field'])

    # For each field in the row, for each type of size, propagate any given sizes upwards.
    for _field, field_data in row_data:
        for size_type in sizes.keys():
            current_size = None
            for size in sizes[size_type]:
                if size in field_data:
                    current_size = field_data[size]
                elif current_size:
                    field_data[size] = current_size

    # For each size option, set the width properties of each field as defined.
    # Default to all 12 for small sizes, and calculate any unspecified widths.
    for size in sizes['field']:
        total = sum(y[size] for x, y in row_data if size in y)
        unspecified = len([x for x, y in row_data if size not in y or not y[size]])
        if unspecified == len(row_data) and size == 'small':
            unspecified_widths = [12] * unspecified
        elif unspecified:
            unspecified_widths = calculate_widths(unspecified, 12 - total)
        for _field, data in row_data:
            if size in data and data[size]:
                continue
            data[size] = unspecified_widths.pop(0)

    # For each field, set the prefix/postfix widths if not defined.
    for _field, field_data in row_data:
        for size_type in sizes.keys():
            if size_type == 'field':
                continue
            for size in sizes[size_type]:
                if size not in field_data or not field_data[size]:
                    field_data[size] = 3

class FoundationFormNode(template.Node):
    '''
    A template node for a form field capabale of rendering Foundation-specific markup.
    '''

    def __init__(self, form, nodelist, csrf_enabled=TrueFilterExpression, collapse_container=FalseFilterExpression):
        self.form = form
        self.nodelist = nodelist
        self.csrf_enabled = csrf_enabled
        self.collapse_container = collapse_container

    def render(self, context):
        # pylint: disable=R0912
        # pylint: disable=R0914
        form = self.form.resolve(context)
        form_template = get_template('nuit/includes/_form.html')

        layout_instructions = []
        for line in self.nodelist.render(context).splitlines():
            if not line.strip():
                continue
            row_data = []
            for field_data in line.split(';'):
                field_data = field_data.strip()
                if not field_data:
                    continue
                try:
                    field_name, data = [x.strip() for x in field_data.split(' ', 1)]
                except ValueError:
                    row_data.append((field_data, {}))
                else:
                    try:
                        row_data.append((field_name, literal_eval(data)))
                    except SyntaxError:
                        raise template.TemplateSyntaxError('Invalid parameters for field %s' % field_name)

            normalise_row(row_data)
            layout_instructions.append(row_data)

        all_fields = [field.name for field in form.visible_fields()]

        field_layout = []
        for field_line in layout_instructions:
            fields = []
            for field, data in field_line:
                try:
                    fields.append(FoundationFormField(form[field], **data))
                    all_fields.remove(field)
                except KeyError:
                    raise template.TemplateSyntaxError('Field %s not found in form' % field)
            field_layout.append(fields)

        if all_fields:
            remaining_fields = []
            for field in all_fields:
                remaining_fields.append(FoundationFormField(form[field]))
            field_layout.append(remaining_fields)

        context.update({
            'form': form,
            'fields': field_layout,
            'csrf_enabled': self.csrf_enabled.resolve(context),
            'collapse_container': self.collapse_container.resolve(context),
        })

        # Render the form. Template expects the form object, and a fields object
        # which is a list of list, each list representing a row of the form.
        # Each element in a row list is a FoundationFormField object.

        try:
            return form_template.render(context)
        finally:
            context.pop()

@register.tag
def foundation_form(parser, token):
    '''
    Renders a form with Foundation markup.
    '''
    bits = token.split_contents()
    if len(bits) < 2:
        raise template.TemplateSyntaxError('Incorrect number of arguements for %s, Form object required' % bits[0])
    form = parser.compile_filter(bits[1])
    kwargs = token_kwargs(bits[2:], parser)
    nodelist = parser.parse(('end_foundation_form',))
    parser.delete_first_token()
    return FoundationFormNode(form, nodelist, **kwargs)


@register.simple_tag
def query_transform(request, **kwargs):
    query = request.GET.copy()
    for k, v in kwargs.iteritems():
        query[k] = v
    return query.urlencode()



from django.views.generic import ListView
from django.db.models import Q
from six import string_types

class SearchableListView(ListView):
    '''
    Render a list of objects that's searchable.
    '''
    # pylint: disable=R0901
    # pylint: disable=E1101
    #: The fields the search will be performed on, using ``queryset.filter``.
    search_fields = ()

    def get_context_data(self, **kwargs):
        context = super(SearchableListView, self).get_context_data(**kwargs)
        if self.search_fields:
            context['search'] = True
            query = self.request.GET.get('q')
            if query:
                context['search_query'] = query
        return context

    def get_queryset(self):
        queryset = super(SearchableListView, self).get_queryset()
        if self.search_fields:
            query = self.request.GET.get('q')
            if query is None:
                return queryset
            queryset = self.search_queryset(queryset, query)
        return queryset

    def search_queryset(self, queryset, search_term):
        query = Q()
        for field in self.search_fields:
            lookup = 'icontains'
            if not isinstance(field, string_types):
                field, lookup = field
            query = query | Q(**{'%s__%s' % (field, lookup): search_term})
        return queryset.filter(query)



'''Tests for nuit'''
# pylint: disable=R0904
import django
from django.test import TestCase, RequestFactory, TransactionTestCase
from django.test.utils import override_settings
from django.contrib.auth.models import Permission, User
from django.contrib.messages import constants
from django.core.paginator import Paginator
from django.core.urlresolvers import ResolverMatch
from django.template import Template, Context, TemplateSyntaxError
from bs4 import BeautifulSoup as soup
import unittest

from .forms import TestForm
from ..context_processors import nuit as nuit_context_processor
from ..templatetags.nuit import message_class, set_active_menu, menu_item, calculate_widths, pagination_menu
from ..utils import user_can_see_view
from ..views import SearchableListView

from . import views

from django.db import models

import string
from six import iteritems


def setup_view(view, request, *args, **kwargs):
    """Mimic as_view() returned callable, but returns view instance.

    args and kwargs are the same you would pass to ``reverse()``

    """
    view.request = request
    view.args = args
    view.kwargs = kwargs
    return view

class Publisher(models.Model):
    name = models.CharField(max_length=30)
    address = models.CharField(max_length=50)
    city = models.CharField(max_length=60)
    state_province = models.CharField(max_length=30)
    country = models.CharField(max_length=50)
    website = models.URLField()

class NuitViews(TestCase):
    '''Tests Nuit's views'''

    def setUp(self):
        for letter in string.ascii_uppercase:
            Publisher(
                name = 'Publisher %s' % letter,
                address = 'Address %s' % letter,
                city = 'City %s' % letter,
                state_province = 'State %s' % letter,
                website = 'http://www.publisher%s.com' % letter,
            ).save()

    def setup_list_view(self, query_string=None, search_fields=None):
        request = RequestFactory().get('/fake-path/?%s' % query_string)
        view = SearchableListView()
        view.model = Publisher
        view.object_list = []
        if search_fields:
            view.search_fields = search_fields
        return setup_view(view, request)

    def test_searchable_list_view_context_no_fields(self):
        view = self.setup_list_view()
        context = view.get_context_data(object_list=[])
        self.assertTrue('search' not in context)
        self.assertTrue('search_query' not in context)

    def test_searchable_list_view_context_no_search(self):
        view = self.setup_list_view(search_fields=('name', 'address',))
        context = view.get_context_data(object_list=[])
        self.assertTrue(context['search'])

    def test_searchable_list_view_context_search(self):
        view = self.setup_list_view('q=isher%20A', ('name', 'address',))
        context = view.get_context_data(object_list=[])
        self.assertTrue(context['search'])
        self.assertEqual('isher A', context['search_query'])

    def test_searchable_list_view_queryset_no_fields(self):
        view = self.setup_list_view()
        self.assertEqual(Publisher.objects.count(), view.get_queryset().count())

    def test_searchable_list_view_queryset_no_search(self):
        view = self.setup_list_view(search_fields=('name', 'address',))
        self.assertEqual(Publisher.objects.count(), view.get_queryset().count())

    def test_searchable_list_view_queryset_search(self):
        view = self.setup_list_view('q=isher%20A', ('name', 'address',))
        self.assertEqual(1, view.get_queryset().count())

    def test_searchable_list_view_queryset_different_lookup(self):
        view = self.setup_list_view('q=isher%20A', (('name', 'startswith'),))
        self.assertEqual(0, view.get_queryset().count())

class NuitContextProcessors(TestCase):
    '''Tests Nuit's context processors'''

    sample_settings = {
        'NUIT_APPLICATIONS': [{'name': 'Link', 'link': 'http://www.google.com'}],
        'NUIT_GLOBAL_TITLE': 'Title',
        'NUIT_GLOBAL_LINK': 'http://www.google.com',
        'NUIT_LARGE_LOGO': 'logo.png',
        'NUIT_SMALL_LOGO': 'small_logo.png',
        'NUIT_SEARCH_VIEW': 'test',
        'NUIT_SEARCH_PARAMETER': 'q',
        'NUIT_SEARCH_PLACEHOLDER': 'Search me!',
    }

    def test_nuit_context_processor(self):
        with self.settings(**self.sample_settings):
            resulting_dict = nuit_context_processor(None)
        expected_resulting_settings = self.sample_settings
        expected_resulting_settings['NUIT_LOGIN_URL'] = '/accounts/login/'
        expected_resulting_settings['NUIT_LOGOUT_URL'] = '/accounts/logout/'
        self.assertEqual(resulting_dict, expected_resulting_settings)

class NuitHandlers(TestCase):
    '''Tests Nuit's handlers'''
    urls = 'nuit.tests.urls'

    @override_settings(COMPRESS_OFFLINE=False)
    def test_error_400(self):
        response = self.client.get('/error400/')
        self.assertEqual(response.status_code, 400)

    @override_settings(COMPRESS_OFFLINE=False)
    def test_error_403(self):
        response = self.client.get('/error403/')
        self.assertEqual(response.status_code, 403)

    @override_settings(COMPRESS_OFFLINE=False)
    def test_error_404(self):
        response = self.client.get('/error404/')
        self.assertEqual(response.status_code, 404)

    @override_settings(COMPRESS_OFFLINE=False)
    def test_error_500(self):
        response = self.client.get('/error500/')
        self.assertEqual(response.status_code, 500)

class FakeMessage(object):
    def __init__(self, level):
        self.level = level

class NuitTemplateFilters(TestCase):
    '''Tests Nuit's template filters'''

    def test_message_class(self):
        self.assertEqual('alert', message_class(FakeMessage(constants.ERROR)))
        self.assertEqual('info', message_class(FakeMessage(constants.INFO)))
        self.assertEqual('invalid-level', message_class(FakeMessage('invalid-level')))

class NuitTemplateTags(TestCase):
    '''Tests Nuit's template tags'''

    def test_pagination_menu(self):

        for letter in string.ascii_uppercase:
            Publisher(
                name = 'Publisher %s' % letter,
                address = 'Address %s' % letter,
                city = 'City %s' % letter,
                state_province = 'State %s' % letter,
                website = 'http://www.publisher%s.com' % letter,
            ).save()

        paginator = Paginator(Publisher.objects.all(), 2)

        context = pagination_menu({}, paginator.page(1))
        self.assertEqual([1, 2, 3, None, 12, 13], context['page_list'])

        context = pagination_menu({}, paginator.page(7))
        self.assertEqual([1, 2, None, 5, 6, 7, 8, 9, None, 12, 13], context['page_list'])

        context = pagination_menu({}, paginator.page(12))
        self.assertEqual([1, 2, None, 10, 11, 12, 13], context['page_list'])

    def test_set_active_menu(self):
        output = soup(set_active_menu('bob')).find('span')
        self.assertEqual('bob', output.text)
        self.assertTrue('nuit-active-menu' in output.attrs['class'])
        self.assertTrue('hide' in output.attrs['class'])

    def test_menu_item(self):

        output = soup(menu_item({}, link='/', name='bob', always_display=True)).find('li')
        self.assertEqual('bob', output.text)
        self.assertTrue('menu-bob' in output.attrs['class'])
        self.assertEqual('/', output.find('a').attrs['href'])

        output = soup(menu_item({}, link='/', name='bob', id='bobby', current=True, unavailable=True, always_display=True)).find('li')
        self.assertEqual('bob', output.text)
        self.assertTrue('menu-bobby' in output.attrs['class'])
        self.assertTrue('current' in output.attrs['class'])
        self.assertTrue('unavailable' in output.attrs['class'])
        self.assertEqual('/', output.find('a').attrs['href'])

    def test_calculate_widths(self):
        self.assertEqual([4, 4, 4], calculate_widths(3))
        self.assertEqual([3, 3, 3, 3], calculate_widths(4))
        self.assertEqual([2, 2, 2, 2, 4], calculate_widths(5))
        self.assertEqual([1], calculate_widths(1, 1))

    def test_menu_section(self):
        template = Template('''
            {% load nuit %}
            {% menu_section %}
                Section 1
            {% end_menu_section %}
            {% menu_section "Section 2" %}
                Section 2
            {% end_menu_section %}
            {% menu_section title=section_3 is_list=True %}
                <li>Section 3</li>
            {% end_menu_section %}
            {% menu_section link_name="Section 4" id="section-4-id" %}
                Section 4
            {% end_menu_section %}
        '''.strip())
        rendered = soup(template.render(Context({'section_3': 'Section 3'})))
        sections = rendered.findAll('section')

        expected_section_data = [
            {
                'header': False,
                'list': False,
                'attrs': {
                    'id': 'none',
                    'data-link': 'None',
                },
            },
            {
                'header': 'Section 2',
                'list': False,
                'attrs': {
                    'id': 'section-2',
                    'data-link': 'Section 2',
                },
            },
            {
                'header': 'Section 3',
                'list': True,
                'attrs': {
                },
            },
            {
                'header': False,
                'list': False,
                'attrs': {
                    'id': 'section-4-id',
                    'data-link': 'Section 4',
                },
            }
        ]

        self.assertEqual(4, len(sections))

        for section, data in zip(sections, expected_section_data):
            self.assertTrue('right-menu-reveal' in section.attrs['class'])
            if not data['list']:
                self.assertTrue('data-reveal' in section.attrs)
            else:
                self.assertEqual(1, len(section.findAll('nav')))
                self.assertEqual(1, len(section.find('nav').findAll('ul')))
                self.assertTrue('side-nav' in section.find('nav').find('ul').attrs['class'])
            if data['header']:
                self.assertEqual(1, len(section.findAll('h5')))
                self.assertEqual(data['header'], section.find('h5').text)
            for key, value in iteritems(data['attrs']):
                self.assertEqual(value, section.attrs[key])

    def test_app_menu(self):
        template = Template('''
            {% load nuit %}
            {% app_menu %}{% end_app_menu %}
            {% app_menu "Title" %}{% end_app_menu %}
            {% app_menu title %}content{% end_app_menu %}
        '''.strip())
        rendered = soup(template.render(Context({'title': 'Title 2'})))
        sections = rendered.findAll('section')

        self.assertEqual(3, len(sections))

        for section in sections:
            self.assertTrue('main-nav' in section.attrs['class'])
            self.assertEqual(1, len(section.findAll('nav')))
            self.assertEqual(1, len(section.find('nav').findAll('ul')))
            self.assertTrue('side-nav' in section.find('nav').find('ul').attrs['class'])

        self.assertTrue(1, len(sections[1].findAll('h5')))
        self.assertEqual('Title', sections[1].find('h5').text)
        self.assertTrue(1, len(sections[2].findAll('h5')))
        self.assertEqual('Title 2', sections[2].find('h5').text)

def get_soup(template_contents, context_dict):
    return soup(Template(template_contents).render(Context(context_dict)))

class NuitFormTags(TestCase):
    '''Tests Nuit's form rendering - it's complex!'''

    def setUp(self):
        self.form = TestForm()

    def assert_on_widths(self, form_html, expected_fields_and_widths):
        for field, small, medium, large in expected_fields_and_widths:
            element = form_html.find('input', attrs={'name': field})
            parent_row = element.findParent('div', attrs={'class': 'columns'})
            self.assertTrue('small-%d' % small in parent_row.attrs['class'])
            self.assertTrue('medium-%d' % medium in parent_row.attrs['class'])
            self.assertTrue('large-%d' % large in parent_row.attrs['class'])

    def test_empty_form_tag(self):
        # pylint: disable=E1101
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, ((field, 12, 12, 12) for field in self.form.fields))

    def test_some_fields_defined(self):
        # pylint: disable=E1101
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            title

            firstname
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, ((field, 12, 12, 12) for field in self.form.fields))

    def test_fields_on_same_row(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            title; firstname; lastname
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, (
            ('title', 12, 4, 4),
            ('firstname', 12, 4, 4),
            ('lastname', 12, 4, 4),
        ))

    def test_specifying_a_width(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            title {'medium': 2}; firstname;; lastname
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, (
            ('title', 12, 2, 2),
            ('firstname', 12, 5, 5),
            ('lastname', 12, 5, 5),
        ))

    def test_resetting_higher_width(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            title {'small': 2, 'medium': 0}; firstname; lastname
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, (
            ('title', 2, 4, 4),
            ('firstname', 5, 4, 4),
            ('lastname', 5, 4, 4),
        ))

    def test_prefix(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            url {'prefix': 'http://'}
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, (
            ('url', 9, 9, 9),
        ))

    def test_postfix(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            url {'postfix': '.com'}
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, (
            ('url', 9, 9, 9),
        ))

    def test_prefix_custom_widths(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            url {'prefix': 'http://', 'prefix_small': 4, 'prefix_medium': 5, 'prefix_large': 2}
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, (
            ('url', 8, 7, 10),
        ))

    def test_postfix_custom_widths(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form %}
            url {'postfix': '.com', 'postfix_small': 4, 'postfix_medium': 5, 'postfix_large': 2}
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assert_on_widths(form_html, (
            ('url', 8, 7, 10),
        ))

    def test_no_csrf(self):
        form_html = get_soup('''
            {% load nuit %}
            {% foundation_form form csrf_enabled=False %}
            {% end_foundation_form %}
        ''', {'form': self.form})
        self.assertTrue('csrf' not in form_html)

    def test_invalid_field(self):
        with self.assertRaises(TemplateSyntaxError):
            get_soup('''
                {% load nuit %}
                {% foundation_form form %}
                this_field_does_not_exist
                {% end_foundation_form %}
            ''', {'form': self.form})

    def test_invalid_field_data(self):
        with self.assertRaises(TemplateSyntaxError):
            get_soup('''
                {% load nuit %}
                {% foundation_form form %}
                title invalid-field-data!!!; firstname; lastname
                {% end_foundation_form %}
            ''', {'form': self.form})

    def test_invalid_call(self):
        with self.assertRaises(TemplateSyntaxError):
            get_soup('''
                {% load nuit %}
                {% foundation_form %}
                {% end_foundation_form %}
            ''', {'form': self.form})


class NuitMenuItemVisibility(TransactionTestCase):
    '''
        Test the selective visibility of menu_items
    '''

    def setUp(self):
        self.staff_user = User.objects.create(username='staff-user', is_staff=True)
        self.staff_user.user_permissions.add(Permission.objects.get(codename='add_user'))
        self.superuser_user = User.objects.create(username='superuser-user', is_superuser=True)
        self.normal_user = User.objects.create(username='normal-user')

    def run_view_test(self, view, user):
        return user_can_see_view(ResolverMatch(view, [], {}), user)

    def test_decorator_visibility(self):
        self.assertTrue(self.run_view_test(views.view_needing_staff, self.staff_user))
        self.assertFalse(self.run_view_test(views.view_needing_staff, self.normal_user))

        self.assertTrue(self.run_view_test(views.view_needing_add_user, self.staff_user))
        self.assertFalse(self.run_view_test(views.view_needing_add_user, self.normal_user))

        self.assertTrue(self.run_view_test(views.ViewNeedingAddUser.as_view(), self.staff_user))
        self.assertFalse(self.run_view_test(views.ViewNeedingAddUser.as_view(), self.normal_user))

    @unittest.skipIf(django.VERSION < (1, 9, 0), 'PermissionRequiredMixin was introduced in 1.9.0')
    def test_mixin_visibility(self):
        '''Test visibility for PermissionRequiredMixin'''
        self.assertTrue(self.run_view_test(views.MixinViewNeedingAddUser.as_view(), self.staff_user))
        self.assertFalse(self.run_view_test(views.MixinViewNeedingAddUser.as_view(), self.normal_user))



'''Nuit autoconfig'''
import django


SETTINGS = {
    # For use with django-csp
    'CSP_STYLE_SRC': (
        "'self'",
        # modernizr.js
        "'sha256-CwE3Bg0VYQOIdNAkbB/Btdkhul49qZuwgNCMPgNY5zw='",
        "'sha256-LpfmXS+4ZtL2uPRZgkoR29Ghbxcfime/CsD/4w5VujE='",
        "'sha256-MZKTI0Eg1N13tshpFaVW65co/LeICXq4hyVx6GWVlK0='",
        "'sha256-YJO/M9OgDKEBRKGqp4Zd07dzlagbB+qmKgThG52u/Mk='",
        # jquery.js
        "'sha256-47DEQpj8HBSa+/TImW+5JCeuQeRkm5NMpJWZG3hSuFU='",
    ),
    'TEMPLATE_CONTEXT_PROCESSORS': [
        'django.core.context_processors.request',
        'nuit.context_processors.nuit',
    ],
    'INSTALLED_APPS': [
        'django.contrib.staticfiles',
        'foundation_scss',
        'foundation_icons',
        'jquery',
        'bourbon',
        'pipeline',
    ],
    'PIPELINE': {
        'COMPILERS': (
            'pipeline.compilers.sass.SASSCompiler',
        ),
        'CSS_COMPRESSOR': None,
        'SASS_ARGUMENTS': '--quiet',
        'STYLESHEETS': {
            'nuit': {
                'source_filenames': (
                  'nuit.scss',
                ),
                'output_filename': 'nuit.css',
            },
        },
    },
    'STATICFILES_FINDERS': [
        'pipeline.finders.PipelineFinder',
    ],
    'STATICFILES_STORAGE': 'pipeline.storage.PipelineStorage',
    'TEMPLATES': [
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.request',
                    'nuit.context_processors.nuit',
                ],
            },
        },
    ],
}

if django.VERSION >= (1,9):
    SETTINGS['TEMPLATES'][0]['OPTIONS']['builtins'] = [
        'nuit.templatetags.nuit',
    ]

# We need to add this globally as we're making a new ExtendsNode
# and this needs to be the first node in the template.
try:
    from django.template.base import add_to_builtins
except ImportError:
    pass
else:
    add_to_builtins('nuit.templatetags.nuit')



'''Nuit Status Code Handlers'''
from django.shortcuts import render

def generic_handler(request, template, status, context=None):
    '''
    Return a response with a particular status code, rendering a template with a specified context.
    '''
    context = context or {}
    return render(request, template, context, status=status)

def handler500(request):
    '''
    View handling execeptions, to be used as :data:`django.conf.urls.handler500`
    '''
    return generic_handler(request, 'nuit/generic/500.html', 500)

def handler400(request):
    '''
    View handling bad requests, to be used as :data:`django.conf.urls.handler400`
    '''
    return generic_handler(request, 'nuit/generic/400.html', 400)

def handler403(request):
    '''
    View handling permission denied exceptions, to be used as :data:`django.conf.urls.handler403`
    '''
    return generic_handler(request, 'nuit/generic/403.html', 403)

def handler404(request):
    '''
    View handling invalid URLs, to be used as :data:`django.conf.urls.handler404`
    '''
    return generic_handler(request, 'nuit/generic/404.html', 404)



'''Context Processors for Nuit'''
from django.conf import settings

def nuit(request):

    context = {
        'NUIT_APPLICATIONS': getattr(settings, 'NUIT_APPLICATIONS', None),
        'NUIT_GLOBAL_TITLE': getattr(settings, 'NUIT_GLOBAL_TITLE', None),
        'NUIT_GLOBAL_LINK': getattr(settings, 'NUIT_GLOBAL_LINK', None),
        'NUIT_LARGE_LOGO': getattr(settings, 'NUIT_LARGE_LOGO', None),
        'NUIT_SMALL_LOGO': getattr(settings, 'NUIT_SMALL_LOGO', None),
        'NUIT_LOGOUT_URL': getattr(settings, 'LOGOUT_URL', None),
        'NUIT_LOGIN_URL': getattr(settings, 'LOGIN_URL', None),
        'NUIT_SEARCH_VIEW': getattr(settings, 'NUIT_SEARCH_VIEW', None),
        'NUIT_SEARCH_PLACEHOLDER': getattr(settings, 'NUIT_SEARCH_PLACEHOLDER', None),
        'NUIT_SEARCH_PARAMETER': getattr(settings, 'NUIT_SEARCH_PARAMETER', None),
    }

    return context



'''setup.py for nuit'''
from setuptools import setup, find_packages
from nuit.version import __VERSION__

setup(
    name                 = 'django-nuit',
    version              = __VERSION__,
    description          = 'Django Netnix User Interface Tools',
    author               = 'Ben Cardy',
    author_email         = 'ben.cardy@ocado.com',
    maintainer           = 'Netnix Team',
    maintainer_email     = 'netnix@ocado.com',
    packages             = find_packages(),
    install_requires     = [
        'django-foundation-statics >= 5.4.7-2',
        'django-jquery',
        'django-foundation-icons>=3.1',
        'django < 1.10',
        'django-autoconfig',
        'django-bourbon',
        'django-pipeline >= 1.6.0',
        'six'
    ],
    include_package_data = True,
    test_suite           = 'setuptest.setuptest.SetupTestSuite',
    tests_require        = ['beautifulsoup4', 'django-setuptest'],
)



'''Utils.'''

from django.core.exceptions import PermissionDenied
import six
import inspect


def _closure(fn):
    return six.get_function_closure(fn)


def _code(fn):
    return six.get_function_code(fn)


def _func(fn):
    return six.get_method_function(fn)


def has_closure(fn):
    return hasattr(fn, 'func_closure') or hasattr(fn, '__closure__')


def has_code(fn):
    return inspect.isfunction(fn)


def get_callable_cells(function):
    '''
    Iterate through all of the decorators on this function,
    and put those that might be callable onto our callable stack.
    '''
    callables = []
    if not has_closure(function):
        if hasattr(function, 'view_func'):
            return get_callable_cells(function.view_func)
    if not _closure(function):
        return [function]
    for closure in _closure(function):
        if hasattr(closure.cell_contents, '__call__'):
            # Class-based view does not have a .func_closure attribute.
            # Instead, we want to look for decorators on the dispatch method.
            # We can also look for decorators on a "get" method, if one exists.
            if hasattr(closure.cell_contents, 'dispatch') and inspect.ismethod(closure.cell_contents.dispatch):
                callables.extend(get_callable_cells(_func(closure.cell_contents.dispatch)))
                if hasattr(closure.cell_contents, 'get') and inspect.ismethod(closure.cell_contents.get):
                    callables.extend(get_callable_cells(_func(closure.cell_contents.get)))
            elif has_closure(closure.cell_contents) and _closure(closure.cell_contents):
                callables.extend(get_callable_cells(closure.cell_contents))
            else:
                callables.append(closure.cell_contents)
    return [function] + callables


def get_class_based_views(callable_cells):
    '''Find class based views for a set of cells.'''
    for cell in callable_cells:
        if has_closure(cell) and _closure(cell):
            closure_dict = dict(zip(_code(cell).co_freevars, _closure(cell)))
            if 'cls' in closure_dict:
                klass = closure_dict['cls'].cell_contents
                if hasattr(klass, 'dispatch'):
                    yield klass


def get_cbv_dispatch_tests(callable_cells):
    '''Generate tests for class based view dispatch methods.'''
    from django.test import RequestFactory

    for klass in get_class_based_views(callable_cells):
        def test_func(user):
            test_view_class = type('TestView', (klass,), {
                'http_method_names': [],
            })
            request = RequestFactory().get('/')
            request.user = user
            test_view_class.as_view()(request)
            return True
        yield test_func


def get_user_tests(function):
    '''
    Get a list of callable cells attached to this function that have the first
    parameter named "u" or "user".
    '''
    callable_cells = get_callable_cells(function)
    return [
        x for x in callable_cells
        if has_code(x) and (
            _code(x).co_varnames[0] in ["user", "u"] or
            (len(_code(x).co_varnames) > 1 and _code(x).co_varnames[0] in ['self', 'cls'] and _code(x).co_varnames[1] in ['u', 'user'])
        )
    ] + list(get_cbv_dispatch_tests(callable_cells))


def test_view(test, urlconf, user):
    '''
    Run a view test. Add in *args, **kwargs if appropriate.
    '''
    args = [] if 'args' not in _code(test).co_varnames else urlconf.args
    kwargs = {} if 'kwargs' not in _code(test).co_varnames else urlconf.kwargs
    return test(user, *args, **kwargs)


def user_can_see_view(view, user):
    '''
    Can the user see this view?
    '''
    try:
        can_view = all([test_view(test, view, user) for test in get_user_tests(view.func)])
    except PermissionDenied:
        return False

    return can_view
